from __future__ import print_function
"""dataClay Tool code for Python.

The code in this package is intended to be used by the "dataClay Tool", for all
the Python specific stuff.
"""

# TODO The below code should be deprecated as soon as the dclayTool.sh is done

from dataclay.api import init_connection, finish
from dataclay.commonruntime.ExecutionGateway import ExecutionGateway
import grpc
from importlib import import_module
from jinja2 import Template
import os
import sys
from uuid import UUID

from dataclay.commonruntime.Runtime import getRuntime
from dataclay.communication.grpc.messages.common.common_messages_pb2 import LANG_PYTHON
from dataclay.util.StubUtils import deploy_stubs
from dataclay.util.StubUtils import prepare_storage
from dataclay.util.tools.python.PythonMetaClassFactory import MetaClassFactory
from dataclay.util.YamlParser import dataclay_yaml_dump, dataclay_yaml_load

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2017 Barcelona Supercomputing Center (BSC-CNS)'

USAGE_TEXT = """
Current commands:

    ./manage.py createuser <admin_user> <admin_password> <new_user> <new_user_password>
    
Creates a user. Needs to be provided admin credentials in order to perform
the operation.

    ./manage.py registertodataclaypubliccontract <user> <password>
    
Prepares a contract to the dataClay public classes (aka dc_classes, aka contrib
classes).  If the namespace doesn't exist, it will be created.

    ./manage.py registermodule <namespace> <module_name> <user> <password>

Register a whole "model" folder, in which there are multiple classes and packages.
If the namespace doesn't exist, it will be created.

    ./manage.py __new__registermodel <namespace> <python_path> <user> <password>
    
Registers all StorageObject classes in a certain module (relative to the
current folder). The classes are registered onto the provided namespace
(created if non-existent).

A full default contract for all registered classes is created and its 
ContractID is returned.

The user credentials should be provided (a "registrator" kind of user).

    ./manage.py registerdataset <dataset> <user> <password>

Creates a dataset and also its matching datacontract. The creation user and
usage user is the same (provided user and password).

    ./manage.py getstubs <contract_ids> <path> <user> <password>
    
Given a ContractID (or a comma-separated list of ContractIDs) download its 
stubs into the provided path. The user should already have a contract to 
access those classes.
"""


def _execute_from_command_line(argv=None):
    client = getRuntime().ready_clients["@LM"]

    if len(argv) < 2:
        print("You should provide a command to the tool.", file=sys.stderr)
        print(USAGE_TEXT, file=sys.stderr)
        return

    # ToDo: A smarter strategy should be used here. At the moment, all commands
    # ToDo: are hardcoded and behaviour is programmed somewhat ad hoc.
    if argv[1] == "createuser":
        # Create a NORMAL_ROLE user.

        admin_id = client.get_account_id(argv[2])
        admin_credential = (None, argv[3])
        username = argv[4]
        password = argv[5]

        yaml_request_template = Template("""
---
 - !!es.bsc.dataclay.util.management.accountmgr.Account
   username: {{ username }}
   credential: !!es.bsc.dataclay.util.management.accountmgr.PasswordCredential
     password: {{ password }}
   role: NORMAL_ROLE
""")

        yaml_request = yaml_request_template.render(
            username=username,
            password=password,
        )

        client.perform_set_of_new_accounts(admin_id,
                                           admin_credential,
                                           yaml_request)

    elif argv[1] == "registertodataclaypubliccontract":
        username = argv[2]
        credential = (None, argv[3])

        user_id = client.get_account_id(username)

        #########################################################################
        # First, we prepare the namespace:

        try:
            namespace_id = client.get_namespace_id(user_id, credential, "dc_classes")
        except Exception:
            yaml_request = """
---
{namespace}: !!es.bsc.dataclay.util.management.namespacemgr.Namespace
    providerAccountName: {consumer_name}
    name: {namespace_name}
    language: LANG_PYTHON
""".format(consumer_name=username,
           namespace="dc_classes",
           namespace_name="dc_classes")

            yaml_response = client.perform_set_of_operations(user_id, credential, yaml_request)
            response = dataclay_yaml_load(yaml_response)

            namespace_id = response["namespaces"]["dc_classes"]

        #########################################################################
        # Then we prepare the classes
        from dataclay import contrib
        modules = contrib.MODULES_TO_REGISTER

        mfc = MetaClassFactory(namespace="dc_classes",
                               responsible_account=username)

        for m_str in modules:
            m = import_module("dataclay.contrib.%s" % m_str)

            for c_str in getattr(m, "CLASSES_TO_REGISTER"):
                mfc.add_class(getattr(m, c_str))

        client = getRuntime().ready_clients["@LM"]
        result = client.new_class(user_id, LANG_PYTHON, mfc.classes)

        if not result:
            raise RuntimeError("No classes successfully registered --cannot continue")

        class_interface_template = Template("""
{{ brief_name }}interface: &{{ brief_name }}iface !!es.bsc.dataclay.util.management.interfacemgr.Interface
  providerAccountName: {{ username }}
  namespace: dc_classes
  classNamespace: dc_classes
  className: {{ class_name }}
  propertiesInIface: !!set {% if class_info.properties|length == 0 %} { } {% endif %}
  {% for property in class_info.properties %}
    ? {{ property.name }}
  {% endfor %}
  operationsSignatureInIface: !!set {% if class_info.operations|length == 0 %} { } {% endif %}
  {% for operation in class_info.operations %}
    ? {{ operation.nameAndDescriptor }}
  {% endfor %}
""")

        class_interface_in_contract_template = Template("""
    - !!es.bsc.dataclay.util.management.contractmgr.InterfaceInContract
      iface: *{{ brief_name }}iface
      implementationsSpecPerOperation: !!set {% if class_info.operations|length == 0 %} { } {% endif %}
        {% for operation in class_info.operations %}
          ? !!es.bsc.dataclay.util.management.contractmgr.OpImplementations
            operationSignature: {{ operation.nameAndDescriptor }}
            numLocalImpl: 0
            numRemoteImpl: 0
        {% endfor %}
    """)

        classes_render = list()
        incontract_render = list()
        for class_name, class_info in result.items():
            brief_name = class_name.rsplit(".", 1)[-1].lower()
            classes_render.append(class_interface_template.render(
                username=username,
                brief_name=brief_name,
                class_name=class_name,
                class_info=class_info))
            incontract_render.append(class_interface_in_contract_template.render(
                brief_name=brief_name, class_info=class_info
            ))

        yaml_request_template = Template("""
---
{% for class_iface in class_interfaces %}
{{ class_iface }}
{% endfor %}
contribcontract: !!es.bsc.dataclay.util.management.contractmgr.Contract
  beginDate: 2012-09-10T20:00:03
  endDate: 2020-09-10T20:00:04
  namespace: dc_classes
  providerAccountID: {{ user_id }}
  applicantsAccountsIDs:
    ? {{ user_id }}
  interfacesInContractSpecs:
    {% for contract in contracts %}
    {{ contract }}
    {% endfor %}
  publicAvailable: True
""")
        yaml_request = yaml_request_template.render(
            user_id=user_id,
            class_interfaces=classes_render,
            contracts=incontract_render
        )

        yaml_response = client.perform_set_of_operations(user_id, credential, yaml_request)
        response = dataclay_yaml_load(yaml_response)

        #########################################################################
        # Now (hopefully) the contract for the public classes has been obtained
        print(" ===> The ContractID for the registered classes is:", file=sys.stderr)
        print(response["contracts"]["contribcontract"], file=sys.stderr)

    elif argv[1] == "__new__registermodel":
        namespace = argv[2]
        python_path = argv[3]
        username = argv[4]
        password = argv[5]

        # Ugly stuff related to the namespace first...
        credential = (None, password)
        user_id = client.get_account_id(username)

        yaml_request_template = Template("""
---
{{ namespace }}: !!es.bsc.dataclay.util.management.namespacemgr.Namespace
  providerAccountName: {{ username }}
  name: {{ namespace }}
  language: LANG_PYTHON
""")
        yaml_request = yaml_request_template.render(
            namespace=namespace,
            username=username
        )

        try:
            client.perform_set_of_operations(user_id, credential, yaml_request)
        except grpc.RpcError:
            # We assume that the namespace already exists
            pass

        # Then use the new register_model shiny stuff
        from .functions import register_model
        register_model(namespace=namespace,
                       python_path=python_path,
                       username=username,
                       password=password)

    elif argv[1] == "registermodule":
        namespace = argv[2]
        module_name = argv[3]
        username = argv[4]
        credential = (None, argv[5])

        user_id = client.get_account_id(username)

        yaml_request_template = Template("""
---
{{ namespace }}: !!es.bsc.dataclay.util.management.namespacemgr.Namespace
  providerAccountName: {{ username }}
  name: {{ namespace }}
  language: LANG_PYTHON
""")
        yaml_request = yaml_request_template.render(
            namespace=namespace,
            username=username
        )

        try:
            client.perform_set_of_operations(user_id, credential, yaml_request)
        except grpc.RpcError:
            # We assume that the namespace already exists
            pass

        mfc = MetaClassFactory(namespace=namespace,
                               responsible_account=username)

        # Scrap the classes in the module
        registered_classes = list()
        module = import_module(module_name)
        for thing_name in dir(module):
            thing = getattr(module, thing_name)

            if not isinstance(thing, ExecutionGateway):
                continue

            # Thing seems to be a DataClayObject class
            if thing.__module__ != module_name:
                print("The module for %s is %s, ignoring because it does not equals %s" % (
                    thing, thing.__module__, module_name
                    ), file=sys.stderr)
                continue

            # Ok, that's a valid class
            mfc.add_class(thing)
            registered_classes.append(thing.__name__)

        registrator_id = client.get_account_id(username)
        result = client.new_class(registrator_id,
                                  credential,
                                  LANG_PYTHON,
                                  mfc.classes)
        
        print("Was gonna register: %s\nEventually registered: %s" % (
            registered_classes, result.keys()), file=sys.stderr)

        if len(result.keys()) == 0:
            print("No classes registered, exiting", file=sys.stderr)
            return

        interfaces = list()
        interfaces_in_contract = list()

        for class_name, class_info in result.items():
            ref_class_name = class_name.replace('.', '')

            interfaces.append(Template("""
{{ class_name }}interface: &{{ ref_class_name }}iface !!es.bsc.dataclay.util.management.interfacemgr.Interface
  providerAccountName: {{ username }}
  namespace: {{ namespace }}
  classNamespace: {{ namespace }}
  className: "{{ reg_class_name }}"
  propertiesInIface: !!set {% if class_info.properties|length == 0 %} { } {% endif %}
  {% for property in class_info.properties %}
     ? {{ property.name }}
  {% endfor %}
  operationsSignatureInIface: !!set {% if class_info.operations|length == 0 %} { } {% endif %}
  {% for operation in class_info.operations %} 
    ? {{ operation.nameAndDescriptor }}
  {% endfor %}
""").render(
                class_name=ref_class_name,
                reg_class_name=class_name,
                username=username,
                namespace=namespace,
                class_info=class_info))

            interfaces_in_contract.append(Template("""
    - !!es.bsc.dataclay.util.management.contractmgr.InterfaceInContract
      iface: *{{ class_name }}iface
      implementationsSpecPerOperation: !!set {% if class_info.operations|length == 0 %} { } {% endif %}
        {% for operation in class_info.operations %}
          ? !!es.bsc.dataclay.util.management.contractmgr.OpImplementations
            operationSignature: {{ operation.nameAndDescriptor }}
            numLocalImpl: 0
            numRemoteImpl: 0
        {% endfor %}
""").render(
                class_name=ref_class_name,
                class_info=class_info))

        contract = Template("""
{{ namespace }}contract: !!es.bsc.dataclay.util.management.contractmgr.Contract
  beginDate: 1980-01-01T00:00:01
  endDate: 2055-12-31T23:59:58
  namespace: {{ namespace }}
  providerAccountID: {{ 00000000-00000000-00000000-00000000-00000000 }}
  applicantsNames:
    ? {{ username }}
  interfacesInContractSpecs:
{{ interfaces_in_contract }}
  publicAvailable: True
""").render(
            namespace=namespace,
            username=username,
            interfaces_in_contract="\n".join(interfaces_in_contract)
        )

        yaml_request = "\n".join(interfaces) + contract
        print(" ===> The yaml for performing is %s" % (yaml_request), file=sys.stderr)

        yaml_response = client.perform_set_of_operations(user_id, credential, yaml_request)
        response = dataclay_yaml_load(yaml_response)

        print(" ===> The ContractID for the registered classes is: ", file=sys.stderr)
        print(response["contracts"]["%scontract" % namespace])

    elif argv[1] == "registerdataset":
        dataset = argv[2]
        username = argv[3]
        credential = (None, argv[4])

        user_id = client.get_account_id(username)

        yaml_request_template = Template("""
{{ dataset }}: !!es.bsc.dataclay.util.management.datasetmgr.DataSet
  dataClayID: {{ 11111111-00000000-00000000-00000000-00000000 }}
  providerAccountID: {{ 00000000-00000000-00000000-00000000-00000000 }}
  name: {{ dataset }}

{{ dataset }}datacontract: !!es.bsc.dataclay.util.management.datacontractmgr.DataContract
  beginDate: 1980-01-01T00:00:01
  endDate: 2055-12-31T23:59:58
  providerAccountID: {{ 00000000-00000000-00000000-00000000-00000000 }}
  providerDataSetID: {{ 11111111-00000000-00000000-00000000-00000000 }}
  applicantsNames:
    ? {{ username }}
  publicAvailable: True
""")
        yaml_request = yaml_request_template.render(
            dataset=dataset,
            username=username
        )

        try:
            client.perform_set_of_operations(user_id, credential, yaml_request)
        except grpc.RpcError as e:
            print("Tried to do that operation and received: %s" % e , file=sys.stderr)

    elif argv[1] == "getstubs":
        # TODO: If this part is still used, check that contract_ids should be a list.
        contract_ids = map(UUID, argv[2].split(','))
        path = argv[3]
        username = argv[4]
        credential = (None, argv[5])

        user_id = client.get_account_id(username)

        prepare_storage(path)

        babel_data = client.get_babel_stubs(user_id,
                                            credential,
                                            contract_ids)

        with open(os.path.join(path, "babelstubs.yml"), 'wb') as f:
            f.write(babel_data)

        all_stubs = client.get_stubs(user_id, credential,
                                     LANG_PYTHON,
                                     contract_ids)

        for key, value in all_stubs.items():
            with open(os.path.join(path, key), 'wb') as f:
                f.write(value)

        deploy_stubs(path)
    else:
        print("Unknown command." , file=sys.stderr)
        print(USAGE_TEXT , file=sys.stderr)
        return


def execute_from_command_line(argv=None):
    """Given the calling arguments to the manage.py script, do stuff.

    :param argv: Typically, sys.argv. Should be explicitly set by caller.
    :return: Nothing.
    """
    # Perform implicit initialization of connections (client.properties only, no storage.properties)
    client_properties_path = os.getenv("DATACLAYCLIENTCONFIG", "./cfgfiles/client.properties")
    assert client_properties_path, "dataclay.tool module can only be called with DATACLAYCLIENTCONFIG set"
    init_connection(client_properties_path)

    _execute_from_command_line(argv)

    # Do the cleanup to avoid __del__ messages of gRPC library
    finish()




""" Class description goes here. """

'''
Created on 19 abr. 2018

@author: dgasull
'''
from dataclay.serialization.python.lang.IntegerWrapper import IntegerWrapper
from dataclay.serialization.python.lang.BooleanWrapper import BooleanWrapper
from dataclay.serialization.python.lang.StringWrapper import StringWrapper
import logging 

logger = logging.getLogger("ReferenceCounting")


class ReferenceCounting(object):
    '''
    classdocs+
    '''

    def __init__(self):
        '''
        Constructor
        '''
        self.reference_counting = dict()

    def increment_reference_counting(self, oid, hint):
        
        if not hint in self.reference_counting: 
            references_per_hint = dict()
            self.reference_counting[hint] = references_per_hint
        else:
            references_per_hint = self.reference_counting.get(hint)
        
        if oid in references_per_hint:
            num_refs = references_per_hint.get(oid) + 1
        else:
            num_refs = 1
        references_per_hint[oid] = num_refs  
        
    def serialize_reference_counting(self, referrer_oid, io_file):
        """ TODO: IMPORTANT: this should be removed in new serialization by using paddings to directly access reference counters inside
         metadata. """
        """
        @postcondition: Serialize reference counting (garbage collector information)
        @param referrer_oid: ID of referrer object
        @param io_file: Buffer in which to serialize
        @param reference_counting: Reference counting to serialize
        """
                
        IntegerWrapper().write(io_file, len(self.reference_counting))
        for location, ref_counting_in_loc in self.reference_counting.items():
            if location is None:
                BooleanWrapper().write(io_file, True)
            else:
                BooleanWrapper().write(io_file, False)
                StringWrapper().write(io_file, str(location))
            
            IntegerWrapper().write(io_file, len(ref_counting_in_loc))
            for oid, counter in ref_counting_in_loc.items():
                StringWrapper().write(io_file, str(oid))
                IntegerWrapper().write(io_file, counter)
                
    def has_no_references(self):
        return len(self.reference_counting) == 0



from __future__ import print_function
"""Functions that can be called from the command line (dataClay tool related).

See the __main__.py entrypoint for more information on the scenarios in which
the functions here are called.
"""
import atexit
from dataclay.api import init_connection
from jinja2 import Template
import os
import sys
from uuid import UUID

from dataclay.communication.grpc.messages.common.common_messages_pb2 import LANG_PYTHON
from dataclay.util.StubUtils import deploy_stubs
from dataclay.util.StubUtils import prepare_storage
from dataclay.util.YamlParser import dataclay_yaml_dump, dataclay_yaml_load
from dataclay.util.tools.python.PythonMetaClassFactory import MetaClassFactory


def _establish_client():
    client = init_connection(os.getenv("DATACLAYCLIENTCONFIG", "./cfgfiles/client.properties"))
    # The client has been correctly acquired, set the "shutdown hook"
    atexit.register(lambda: client.close())
    return client


def register_model(username, password, namespace, python_path):
    client = _establish_client()

    credential = (None, password)
    user_id = client.get_account_id(username)

    mfc = MetaClassFactory(namespace=namespace,
                           responsible_account=username)

    full_python_path = os.path.abspath(python_path)

    # Load the model_package from the path
    if full_python_path not in sys.path:
        added_to_path = True
        sys.path.insert(0, full_python_path)
    else:
        added_to_path = False

    n_base_skip = len(full_python_path.split(os.sep))

    for dirpath, dirnames, filenames in os.walk(full_python_path):
        base_import = ".".join(
            dirpath.split(os.sep)[n_base_skip:]
        )

        for f in filenames:
            if not f.endswith(".py"):
                continue

            if f == "__init__.py":
                if not base_import:
                    print("Ignoring `__init__.py` at the root of the model folder (do not put classes there!)", file=sys.stderr)
                    continue
                import_str = base_import
            else:
                if not base_import:
                    import_str = os.path.splitext(f)[0]
                else:
                    import_str = "%s.%s" % (base_import, os.path.splitext(f)[0])

            # Try to import
            mfc.import_and_add(import_str)

    result = client.new_class(user_id,
                              credential,
                              LANG_PYTHON,
                              mfc.classes)
    
    print("Was gonna register: %s\nEventually registered: %s" % (mfc.classes, result.keys()), file=sys.stderr)

    if len(result.keys()) == 0:
        print("No classes registered, exiting", file=sys.stderr)
        return

    interfaces = list()
    interfaces_in_contract = list()
    for class_name, class_info in result.items():
        ref_class_name = class_name.replace('.', '')

        interfaces.append(Template("""
{{ class_name }}interface: &{{ ref_class_name }}iface !!es.bsc.dataclay.util.management.interfacemgr.Interface
  providerAccountName: {{ username }}
  namespace: {{ namespace }}
  classNamespace: {{ namespace }}
  className: {{ class_name }}
  propertiesInIface: !!set {% if class_info.properties|length == 0 %} { } {% endif %}
  {% for property in class_info.properties %}
     ? {{ property.name }}
  {% endfor %}
  operationsSignatureInIface: !!set {% if class_info.operations|length == 0 %} { } {% endif %}
  {% for operation in class_info.operations %} 
    ? {{ operation.nameAndDescriptor }}
  {% endfor %}
""").render(
            class_name=class_name,
            ref_class_name=ref_class_name,
            username=username,
            namespace=namespace,
            class_info=class_info))

        interfaces_in_contract.append(Template("""
    - !!es.bsc.dataclay.util.management.contractmgr.InterfaceInContract
      iface: *{{ ref_class_name }}iface
      implementationsSpecPerOperation: !!set {% if class_info.operations|length == 0 %} { } {% endif %}
        {% for operation in class_info.operations %}
          ? !!es.bsc.dataclay.util.management.contractmgr.OpImplementations
            operationSignature: {{ operation.nameAndDescriptor }}
            numLocalImpl: 0
            numRemoteImpl: 0
        {% endfor %}
""").render(
            class_name=class_name,
            ref_class_name=ref_class_name,
            class_info=class_info))

    contract = Template("""
{{ namespace }}contract: !!es.bsc.dataclay.util.management.contractmgr.Contract
  beginDate: 1980-01-01T00:00:01
  endDate: 2055-12-31T23:59:58
  namespace: {{ namespace }}
  providerAccountID: {{ user_id }}
  applicantsAccountsIDs:
    ? {{ user_id }}
  interfacesInContractSpecs:
{{ interfaces_in_contract }}
  publicAvailable: True
""").render(
        user_id=user_id,
        namespace=namespace,
        interfaces_in_contract="\n".join(interfaces_in_contract)
    )

    yaml_request = "\n".join(interfaces) + contract

    yaml_response = client.perform_set_of_operations(user_id, credential, yaml_request)
    response = dataclay_yaml_load(yaml_response)

    print(" ===> The ContractID for the registered classes is:", file=sys.stderr)
    print(response["contracts"]["%scontract" % namespace])
    
    # Remove from sys.path the model
    if added_to_path:
        sys.path.remove(full_python_path)
    
    return str(response["contracts"]["%scontract" % namespace])  # For mock testing


def get_stubs(username, password, contract_ids_str, path):
    client = _establish_client()

    try:
        contracts = list(map(UUID, contract_ids_str.split(',')))
    except ValueError:
        raise ValueError("This is not a valid list of contracts: %s" % contract_ids_str)
    credential = (None, password)

    user_id = client.get_account_id(username)
    prepare_storage(path)

    babel_data = client.get_babel_stubs(
        user_id, credential, contracts)

    with open(os.path.join(path, "babelstubs.yml"), 'wb') as f:
        f.write(babel_data)

    all_stubs = client.get_stubs(
        user_id, credential, LANG_PYTHON, contracts)

    for key, value in all_stubs.items():
        with open(os.path.join(path, key), 'wb') as f:
            f.write(value)

    deploy_stubs(path)




""" Class description goes here. """

from dataclay.util.MgrObject import ManagementObject


class AccessedProperty(ManagementObject):
    _fields = [
               "id",  # sql id, internals
               "namespace",
               "className",
               "name",
               ]

    _internal_fields = ["propertyID",
                        ]





""" Class description goes here. """

"""Interceptor for GRPC client calls."""

from grpc import UnaryUnaryClientInterceptor
from logging import getLogger

from dataclay import PrvManager
from dataclay.communication.grpc.paraver import HEADER_CLIENTPORT
from dataclay.paraver.prv_traces import TraceType
from dataclay.commonruntime.Settings import settings

from . import HEADER_MESSAGEID
import time

logger = getLogger(__name__)


class ParaverClientInterceptor(UnaryUnaryClientInterceptor):

    def __init__(self, origin_hostname, remote_hostaddr, remote_port):
        logger.debug("Initialize ParaverClientInterceptor")
        self.request_id = 42
        self.prv_manager = PrvManager.get_manager()
        self.origin_hostname = origin_hostname
        self.remote_hostaddr = remote_hostaddr
        self.remote_port = remote_port

    def intercept_unary_unary(self, continuation, client_call_details, request):
        """
        Override of the UnaryUnaryClientInterceptor method.
        If paraver is active (activate_paraver_traces) add req_id to metadata
        and try to get the client_port from it after the response, then add the send/receive traces to the traces queue. 

        :param continuation: A function that proceeds with the invocation by executing the next interceptor in chain
                             or invoking the actual RPC on the underlying Channel.
                             It is the interceptor's responsibility to call it if it decides to move the RPC forward.
                             The interceptor can use response_future = continuation(client_call_details, request) to continue with the RPC.
                             continuation returns an object that is both a Call for the RPC and a Future. 
                             In the event of RPC completion, the return Call-Future's result value will be the response message of the RPC.
                             Should the event terminate with non-OK status, the returned Call-Future's exception value will be an RpcError.

        :param client_call_details: A ClientCallDetails object describing the outgoing RPC.
        :param request: The request value for the RPC.
        :return: Response of the RPC.
        """
        if settings.paraver_tracing_enabled:
                
            logger.debug("Intercept")
            req_id = self.request_id
            self.request_id += 1

            metadata = client_call_details.metadata
            if not metadata:
                metadata = list()

            metadata.append((HEADER_MESSAGEID, str(req_id)))

            # It is known that client_call_details is in fact a namedtuple,
            # (i.e. inmutable) so let's take advantage of the _replace method
            new_call_details = client_call_details._replace(metadata=metadata)
            
            send_time = int(time.time() * 1000000000)
            response = continuation(new_call_details, request)
            metadata = response.initial_metadata()

            client_port = dict(metadata).get(HEADER_CLIENTPORT)
            if not client_port:
                # If we can't retrieve the client port assign a static value in client/server interceptor
                client_port = 892892
            
            self.prv_manager.add_network_send(
                send_time,
                TraceType.SEND_REQUEST,
                int(client_port),
                req_id,
                self.remote_hostaddr,
                self.remote_port,
                0,  # unknown/unused message size
                0  # unknown/unused method id
            )

            self.prv_manager.add_network_receive(
                self.remote_hostaddr,
                self.remote_port,
                req_id,
                0
            )
            return response
        else:
            return continuation(client_call_details, request)
            




""" Class description goes here. """

from datetime import datetime
from dataclay.util.ConfigurationFlags import ConfigurationFlags
import logging
import os
import traceback
logger = logging.getLogger('dataclay.api')


class Configuration(object):
    """Configuration management static-ish class."""

    DATE_FORMAT = "%Y-%m-%dT%H:%M:%S"

    # Session control
    ########################
    CHECK_SESSION = False

    # Garbage collector
    ########################

    # Percentage to start flushing objects
    MEMMGMT_PRESSURE_FRACTION = 0.75

    # Percentage to stop flushing objects
    MEMMGMT_EASE_FRACTION = 0.50

    # Number of milliseconds to check if Heap needs to be cleaned.
    MEMMGMT_CHECK_TIME_INTERVAL = 5000

    # Global GC collection interval
    NOCHECK_SESSION_EXPIRATION = datetime.strptime('2120-09-10T20:00:04', DATE_FORMAT)

    # ====================== TIME OUTS AND RETRIES ========================== #

    # NEW DATA-SERVICE OR EXECUTION ENVIRONMENT: WARNING: for each retry there is max_retries_logicmodule
    # Maximum autoregistration retries of a node. 
    MAX_RETRY_AUTOREGISTER = 10
    # Seconds to wait to retry autoregistration of a node. 
    RETRY_AUTOREGISTER_TIME = 3000
        
    # ANY CALL TO LOGICMODULE: clients can wait at init waitForBackends
    # Default value for number of retries in connection to LogicModule. 
    MAX_RETRIES_LOGICMODULE = 5
    # Default value for sleeping before retrying in LM in seconds. 
    SLEEP_RETRIES_LOGICMODULE = 10
        
    # EXECUTION RETRIES
    # Default value for number of retries IN EXECUTION
    MAX_EXECUTION_RETRIES = 5
        
    # WAIT FOR OBJECTS TO BE REGISTERED BY GC
    # Number of millis of time to wait for object to be registered. Default: NO WAIT 
    TIMEOUT_WAIT_REGISTERED = 0  # fixme do we need this?
    # Waiting milliseconds to check if object to be registered. 
    SLEEP_WAIT_REGISTERED = 50
    
    # CHECK ALIVE TIME OUT IN GRPC in seconds
    GRPC_CHECK_ALIVE_TIMEOUT = 120
    
    # Indicates where EE id is stored
    EE_PERSISTENT_INFO_PATH = os.getcwd() + "/"

    # Destination path for traces
    TRACES_DEST_PATH = os.getcwd()

    # How many worker threads should be created/used by ThreadPoolExecutor
    THREAD_POOL_WORKERS = 2

    # Path to Trusted certificates for verifying the remote endpoint's certificate. 
    SSL_CLIENT_TRUSTED_CERTIFICATES = ""

    # Path to identifying certificate for this host
    SSL_CLIENT_CERTIFICATE = ""

    # Path to identifying certificate for this host. 
    SSL_CLIENT_KEY = ""
    
    # Custom header of service alias for calls to Logic module. Used in Traefik.
    LM_SERVICE_ALIAS_HEADERMSG = "logicmodule1"
    
    # Override authority hostname in SSL calls
    SSL_TARGET_AUTHORITY = "proxy"
    
    # Path to state file
    STATE_FILE_PATH = "state.txt"

    # Path to destination traces dir
    TRACES_DEST_PATH = os.getcwd()

    @classmethod
    def read_from_file(cls, flags_path):
        
        # Check if environment variable is set 
        abs_flags_path = os.path.abspath(flags_path)
        print("--- Using global.properties at %s" % abs_flags_path)
        try:
            flags = ConfigurationFlags(property_path=abs_flags_path)
            globalattrs = [a for a in dir(flags) if not a.startswith('__')]
            for attr in globalattrs:
                value = getattr(flags, attr)
                if hasattr(cls, attr):
                    cur_value = getattr(cls, attr)
                    print("Found global.properties variable %s=%s and type %s" % (str(attr), str(value), str(type(cur_value))))
                    setattr(cls, str(attr), type(cur_value)(value))

        except:
            print("No global.properties file found. Using default values")


# Check if environment variable is set 
if "DATACLAYGLOBALCONFIG" in os.environ:
    flags_path = os.environ["DATACLAYGLOBALCONFIG"]
else:
    flags_path = "./cfgfiles/global.properties"
    
Configuration.read_from_file(flags_path)




""" Class description goes here. """

IGNORE_PREFIXES = [
    "es.bsc.dataclay.util.management.classmgr.java",
    "es.bsc.dataclay.util.management.classmgr.features",
]

IGNORE_CLASSES = [
    "es.bsc.dataclay.util.management.classmgr.PrefetchingInformation",
]



from setuptools import setup, find_packages
import os 

def parse_requirements(filename):
    """ load requirements from a pip requirements file """
    lineiter = (line.strip() for line in open(filename))
    return [line for line in lineiter if line and not line.startswith("#")]
# parse_requirements
install_reqs = parse_requirements("./requirements.txt")

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='dataClay',
      version=2.4,
      install_requires=install_reqs,
      description='Python library for dataClay',
      packages=find_packages("src"),
      package_dir={'':'src'},
      package_data={
        # All .properties files are valuable "package data"
        '': ['*.properties'],
      },
      long_description=long_description,
      long_description_content_type="text/markdown",
      url="https://www.bsc.es/dataclay",
      project_urls={
          'Documentation': 'https://pyclay.readthedocs.io/en/latest/',
          'Source': 'https://github.com/bsc-dom/pyclay',
      },
      classifiers=[
          "Development Status :: 5 - Production/Stable",
          "Intended Audience :: Science/Research",
          "Intended Audience :: Developers",
          "License :: OSI Approved :: BSD License",
          "Programming Language :: Python",
          "Topic :: Database :: Database Engines/Servers",
          "Topic :: System :: Distributed Computing",
          "Topic :: Software Development :: Libraries :: Application Frameworks",
      ],
      )





""" Class description goes here. """

"""Placeholder module for the properties.

Note that the files included in this folder are (typically) hard links to the
Java folder (containing the properties' files).

Maybe git (or simply developers) will break this. Be cautious.
"""

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2015 Barcelona Supercomputing Center (BSC-CNS)'




""" Class description goes here. """

from uuid import UUID
from yaml import Loader, Dumper

JAVA_UUID_TAG = u'tag:yaml.org,2002:java.util.UUID'
DATACLAY_ID_PREFIX = u'tag:yaml.org,2002:es.bsc.dataclay.util.ids'


def uuid_representer(dumper, data):
    return dumper.represent_scalar(JAVA_UUID_TAG, str(data))


def uuid_constructor(loader, node):
    value = loader.construct_scalar(node)
    return UUID(value)


Dumper.add_representer(UUID, uuid_representer)
Loader.add_constructor(JAVA_UUID_TAG, uuid_constructor)
Loader.add_multi_constructor(
    DATACLAY_ID_PREFIX,
    # This ignores the tag, as ImplementationID, OperationID, *ID are always
    # used directly as their UUID, not their specific type.
    lambda loader, _, node: uuid_constructor(loader, node))




""" Class description goes here. """

"""Utility methods for gRPC clients/server."""

import logging
import sys
import traceback
import uuid
import six

import dataclay.communication.grpc.messages.common.common_messages_pb2 as common_messages

__author__ = 'Enrico La Sala <enrico.lasala@bsc.es>'
__copyright__ = '2017 Barcelona Supercomputing Center (BSC-CNS)'

logger = logging.getLogger(__name__)


def get_credential(cred):
    """Get the pwdCredential of the corresponding protobuf message.

    :param cred: PasswordCredential/Credential protobuf message.
    :return: Depends on the param can return a Credential protobuf message or a PasswordCredential.
    """
    if cred is common_messages.Credential:
        if cred is None:
            return None
        else:
            return cred.password
    else:
        if cred is None:
            return common_messages.Credential()
        else:
            return common_messages.Credential(password=cred[1])


"""get_msg_id_x.

:param x_id: UUID.
:return: ID protobuf message.
"""


def get_msg_id_account(account_id):
    if account_id is None:

        return common_messages.AccountID()
    else:
        return common_messages.AccountID(
            uuid=str(account_id))


def get_msg_id_contract(contract_id):
    if contract_id is None:

        return common_messages.ContractID()
    else:
        return common_messages.ContractID(
            uuid=str(contract_id))


def get_msg_id_credential(credential_id):
    if credential_id is None:

        return common_messages.CredentialID()
    else:

        return common_messages.CredentialID(
            uuid=str(credential_id))


def get_msg_id_datacontract(dc_id):
    if dc_id is None:

        return common_messages.DataContractID()
    else:
        return common_messages.DataContractID(
            uuid=str(dc_id))


def get_msg_id_dataset(dataset_id):
    if dataset_id is None:

        return common_messages.DataSetID()
    else:
        return common_messages.DataSetID(
            uuid=str(dataset_id))


def get_msg_id_namespace(namespace_id):
    if namespace_id is None:

        return common_messages.NamespaceID()
    else:
        return common_messages.NamespaceID(
            uuid=str(namespace_id))


def get_msg_id_class(class_id):
    if class_id is None:

        return common_messages.MetaClassID()
    else:
        return common_messages.MetaClassID(
            uuid=str(class_id))


def get_msg_id_eca(eca_id):
    if eca_id is None:

        return common_messages.ECAID()
    else:
        return common_messages.ECAID(
            uuid=str(eca_id))


def get_msg_id_event_message(event_message_id):
    if event_message_id is None:

        return common_messages.EventMessageID()
    else:
        return common_messages.EventMessageID(
            uuid=str(event_message_id))


def get_msg_id_event_obj_meet_cond(event_omc_id):
    if event_omc_id is None:

        return common_messages.EventObjsMeetConditionID()
    else:
        return common_messages.EventObjsMeetConditionID(
            uuid=str(event_omc_id))


def get_msg_id_exec_env(ee_id):
    if ee_id is None:

        return common_messages.ExecutionEnvironmentID()
    else:
        return common_messages.ExecutionEnvironmentID(
            uuid=str(ee_id))


def get_msg_id_implem(implem_id):
    if implem_id is None:

        return common_messages.ImplementationID()
    else:
        return common_messages.ImplementationID(
            uuid=str(implem_id))


def get_msg_id_interface(interface_id):
    if interface_id is None:

        return common_messages.InterfaceID()
    else:
        return common_messages.InterfaceID(
            uuid=str(interface_id))


def get_msg_id_meta_class(meta_class_id):
    if meta_class_id is None:

        return common_messages.MetaClassID()
    else:
        return common_messages.MetaClassID(
            uuid=str(meta_class_id))


def get_msg_id_object(object_id):
    if object_id is None:

        return common_messages.ObjectID()
    else:
        return common_messages.ObjectID(
            uuid=str(object_id))


def get_msg_id_operation(operation_id):
    if operation_id is None:

        return common_messages.OperationID()
    else:
        return common_messages.OperationID(
            uuid=str(operation_id))


def get_msg_id_property(property_id):
    if property_id is None:

        return common_messages.PropertyID()
    else:
        return common_messages.PropertyID(
            uuid=str(property_id))


def get_msg_id_qual_reg(qual_reg_id):
    if qual_reg_id is None:

        return common_messages.QualitativeRegistryID()
    else:
        return common_messages.QualitativeRegistryID(
            uuid=str(qual_reg_id))


def get_msg_id_resource(resource_id):
    if resource_id is None:

        return common_messages.ResourceID()
    else:
        return common_messages.ResourceID(
            uuid=str(resource_id))


def get_msg_id_session(session_id):
    if session_id is None:

        return common_messages.SessionID()
    else:
        return common_messages.SessionID(
            uuid=str(session_id))


def get_msg_id_storage_loc(storage_loc_id):
    if storage_loc_id is None:

        return common_messages.StorageLocationID()
    else:
        return common_messages.StorageLocationID(
            uuid=str(storage_loc_id))

        
def get_msg_id_backend_id(backend_id):
    if backend_id is None:

        return common_messages.ExecutionEnvironmentID()
    else:
        return common_messages.ExecutionEnvironmentID(
            uuid=str(backend_id))

def get_msg_id_dc(dataclay_id):
    if dataclay_id is None:
        return common_messages.DataClayInstanceID()
    else:
        return common_messages.DataClayInstanceID(
            uuid=str(dataclay_id)
        )
        
get_msg_options = {'account': get_msg_id_account,
                   'contract': get_msg_id_contract,
                   'credential': get_msg_id_credential,
                   'datacontract': get_msg_id_datacontract,
                   'dataset': get_msg_id_dataset,
                   'namespace': get_msg_id_namespace,
                   'eca': get_msg_id_eca,
                   'event_message': get_msg_id_event_message,
                   'obj_meet_cond': get_msg_id_event_obj_meet_cond,
                   'exec_env': get_msg_id_exec_env,
                   'implem': get_msg_id_implem,
                   'interface': get_msg_id_interface,
                   'meta_class': get_msg_id_meta_class,
                   'object': get_msg_id_object,
                   'operation': get_msg_id_operation,
                   'property': get_msg_id_property,
                   'qual_reg': get_msg_id_qual_reg,
                   'resource': get_msg_id_resource,
                   'session': get_msg_id_session,
                   'storage_loc': get_msg_id_storage_loc,
                   'backend_id': get_msg_id_backend_id,
                   'class': get_msg_id_class,
                   'dataclay_instance': get_msg_id_dc}


def get_id(id_msg):
    """Create the ID based on protobuf message.

    :param id_msg: Common protobuf message with uuid_most and uuid_least.

    :return: UUID based on param.
    """    
    if id_msg is None:
        return None
    elif id_msg.uuid is None:
        return None
    elif id_msg.uuid == "":
        return None
    else:
        return uuid.UUID(id_msg.uuid)


def get_id_from_uuid(msg_str):
    """Create the ID based on id message.

    :param msg_str: string UUID.

    :return: Correct UUID.
    """
    if msg_str is None or msg_str == "":
        return None
    else:
        return uuid.UUID(msg_str)


def get_metadata(metadata):
    """Get Metadata GRPC message from DataclayMetaData.

    :param metadata: DataClayMetaData.
    
    :return: MetaData GRPC message.
    """
    if type(metadata) is common_messages.DataClayObjectMetaData:
       
        oids = dict()
        for k, v in metadata.oids.items():
            oids[k] = get_id(v)
        
        classids = dict()
        for k, v in metadata.classids.items():
            classids[k] = get_id(v)
            
        hints = dict()
        for k, v in metadata.hints.items():
            hints[k] = get_id(v)
        
        num_refs = metadata.numRefs
        
        return oids, classids, hints, num_refs

    else:

        one = dict()
    
        for k, v in metadata[0].items():
            one[k] = get_msg_options['object'](v)
    
        two = dict()
    
        for k, v in metadata[1].items():
            two[k] = get_msg_options['meta_class'](v)

        three = dict()

        for k, v in metadata[2].items():
            three[k] = get_msg_options['exec_env'](v)

        request = common_messages.DataClayObjectMetaData(
            oids=one,
            classids=two,
            hints=three,
            numRefs=metadata[3]
        )
    
        return request


def to_hex(val, nbits):
    """Convert a long to hex.

    :param val: long.

    :param nbits: The number of bit two's complement.

    :return: hex value.
    """
    return hex((val + (1 << nbits)) % (1 << nbits))


def convert_neg(x):
    """Convert negative hex not correctly converted.

    :param x: int.

    :return: Correct int.
    """
    if x > 9223372036854775808:
        x -= 18446744073709551616
    return x


def get_obj_with_data_param_or_return(vol_param_or_ret):
    """
    :param vol_param_or_ret: Could be a protobuf msg or a tuple.

    :return: Could return a deserialized message or a ObjectWithDataParamOrReturn message.
    """
    if type(vol_param_or_ret) is common_messages.ObjectWithDataParamOrReturn:
        oid = get_id(vol_param_or_ret.oid)
        class_id = get_id(vol_param_or_ret.classid)
        mdata = get_metadata(vol_param_or_ret.metadata)
        byte_array = vol_param_or_ret.objbytes

        return oid, class_id, mdata, byte_array
    else:
        response = common_messages.ObjectWithDataParamOrReturn(
            oid=get_msg_options['object'](vol_param_or_ret[0]),
            classid=get_msg_options['meta_class'](vol_param_or_ret[1]),
            metadata=get_metadata(vol_param_or_ret[2]),
            objbytes=vol_param_or_ret[3].getvalue()
        )
        return response


def get_lang_param_or_return(param_or_ret):
    """
    :param param_or_ret: Could be a protobuf msg or a list.

    :return: Could return a deserialized message or a LanguageParamOrReturn message.
    """
    if type(param_or_ret) is common_messages.LanguageParamOrReturn:
        mdata = get_metadata(param_or_ret.metadata)
        byte_array = param_or_ret.objbytes
        return mdata, byte_array

    else:
        request = common_messages.LanguageParamOrReturn(
            metadata=get_metadata(param_or_ret[0]),
            objbytes=param_or_ret[1].getvalue()
        )

        return request


def get_immutable_param_or_return(param_or_ret):
    """
    :param param_or_ret: Could be a protobuf msg or a _io.BytesIO.

    :return: Could return a deserialized message or a ImmutableParamOrReturn message.
    """
    if type(param_or_ret) is common_messages.ImmutableParamOrReturn:
        byte_array = param_or_ret.objbytes
        return byte_array

    else:
        request = common_messages.ImmutableParamOrReturn(
            objbytes=param_or_ret.getvalue()
        )
        return request


def get_persistent_param_or_return(param_or_ret):
    """
    :param param_or_ret: Could be a PersistentParamOrReturn protobuf msg or a list.

    :return: Could return a deserialized message or a PersistentParamOrReturn message.
    """
    if type(param_or_ret) is common_messages.PersistentParamOrReturn:
        oid = get_id(param_or_ret.oid)
        class_id = get_id(param_or_ret.classID)
        hint = get_id(param_or_ret.hint)

        return oid, hint, class_id

    else:
        class_id = None
        hint = None
        oid = param_or_ret[0]

        if param_or_ret[1] is not None:
            hint = param_or_ret[1]

        if param_or_ret[2] is not None:
            class_id = param_or_ret[2]

        request = common_messages.PersistentParamOrReturn(
            oid=get_msg_options['object'](oid),
            hint=get_msg_options['exec_env'](hint),
            classID=get_msg_options['meta_class'](class_id))

        return request


def get_param_or_return(param_or_ret_msg):
    """
    :param param_or_ret_msg: Could be None, a SerializedParametersOrReturn protobuf msg or a list.

    :return: Could return an empty SerializedParametersOrReturn message, a deserialized message 
             or a SerializedParametersOrReturn message.
    """
    if param_or_ret_msg is None:
        # create and return empty serialized message
        return common_messages.SerializedParametersOrReturn()
    elif type(param_or_ret_msg) is common_messages.SerializedParametersOrReturn:
        # create and return deserialized message
        num_params = param_or_ret_msg.numParams
    
        imm_objs = dict()       
        lang_objs = dict()
        vol_objs = dict()
        pers_objs = dict()
    
        for k, v in param_or_ret_msg.immParams.items():
            imm_objs[k] = get_immutable_param_or_return(v)
        
        for k, v in param_or_ret_msg.langParams.items():
            lang_objs[k] = get_lang_param_or_return(v)
        
        for k, v in param_or_ret_msg.volatileParams.items():
            vol_objs[k] = get_obj_with_data_param_or_return(v)
        
        for k, v in param_or_ret_msg.persParams.items():
            pers_objs[k] = get_persistent_param_or_return(v)
            
        return num_params, imm_objs, lang_objs, vol_objs, pers_objs
    elif type(param_or_ret_msg) is list:
        # create and return serialized message
        num_params = param_or_ret_msg[0]
        imm_objs = dict()
        lang_objs = dict()
        vol_objs = dict()
        pers_obj = dict()
        
        for k, v in param_or_ret_msg[1].items():
            imm_objs[k] = get_immutable_param_or_return(v)
        
        for k, v in param_or_ret_msg[2].items():
            lang_objs[k] = get_lang_param_or_return(v)
        
        for k, v in param_or_ret_msg[3].items():
            vol_objs[k] = get_obj_with_data_param_or_return(v)
        
        for k, v in param_or_ret_msg[4].items():
            pers_obj[k] = get_persistent_param_or_return(v)

        request = common_messages.SerializedParametersOrReturn(
                numParams=num_params,
                immParams=imm_objs,
                langParams=lang_objs,
                volatileParams=vol_objs,
                persParams=pers_obj
        )
        
        return request
    else:
        raise TypeError("Param serialized_objs type is wrong. It could be None, SerializedParametersOrReturn protobuf msg or a list.")


def return_empty_message(resp_obs):
    """Return a protobuf empty message.

    :param resp_obs: string msg with UUID.
    """
    resp = common_messages.EmptyMessage()
    resp_obs(resp)


"""
HexByteConversion

Convert a byte string to it's hex representation for output or visa versa.

bye_to_hex converts byte string "\xFF\xFE\x00\x01" to the string "FF FE 00 01"
hex_to_byte converts string "FF FE 00 01" to the byte string "\xFF\xFE\x00\x01"
"""


def bye_to_hex(byte_str):
    """Convert a byte string to it's hex string representation e.g. for output.

    :param byte_str: byte string with UUID.
    :return: hex string msg with UUID.
    """

    # Uses list comprehension which is a fractionally faster implementation than
    # the alternative, more readable, implementation below
    #   
    #    hex = []
    #    for aChar in byteStr:
    #        hex.append( "%02X " % ord( aChar ) )
    #
    #    return ''.join( hex ).strip()        

    return ''.join(["%02X " % ord(x) for x in byte_str]).strip()


def hex_to_byte(hex_str):
    """
    Convert a string hex byte values into a byte string. The Hex Byte values may
    or may not be space separated.
    :param hex_str: hex string msg with UUID.
    :return: byte string msg with UUID.
    """

    # The list comprehension implementation is fractionally slower in this case    
    #
    #    hexStr = ''.join( hexStr.split(" ") )
    #    return ''.join( ["%c" % chr( int ( hexStr[i:i+2],16 ) ) \
    #                                   for i in range(0, len( hexStr ), 2) ] )
 
    bytes_str = []

    hex_str = ''.join(hex_str.split(" "))

    for i in range(0, len(hex_str), 2):
        bytes_str.append(chr(int(hex_str[i:i + 2], 16)))

    return ''.join(bytes_str)


def prepare_exception(message, stack):
    """
    Prepare message + stack for proto messages
    :param message: message of the exception.
    :param stack: stackTrace of the exception.
    :return: final string
    """
    if message is not None:
        return prepare_bytes("\n" + "-"*146 + "\nError Message: " + message + "\n" + "-"*146 + "\nServer " + str(stack) + "\n" + "-"*146)
    else:
        return prepare_bytes("\n" + "-"*146 + "\nError Stack: \n" + "-"*146 + "\nServer " + str(stack) + "\n" + "-"*146)


def return_stack():
    """
    Create the stack of the obtained exception
    :return: string stacktrace.
    """
    exc_type, exc_value, exc_traceback = sys.exc_info()
    lines = traceback.format_exception(exc_type, exc_value, exc_traceback)

    return lines[0] + lines[1]

def prepare_bytes(varz):
    """Create the right bytes var for python2/3.
    :param varx: variable to convert in bytes.
    :return: correct bytes of varz.
    """
    if six.PY2:
        v = bytes(varz)
    elif six.PY3:
        v = bytes(varz, "utf-8")
    return v



""" Class description goes here. """

from ..Implementation import Implementation


class JavaImplementation(Implementation):
    _fields = ["dataClayID"
               ]


""" _fields = ["position",
               "accessedProperties",
               "accessedImplementations",
               "includes",
               "prefetchingInfo",
               "reqQuantitativeFeatures",
               "reqQualitativeFeatures",
               "namespace",
               "className",
               "opNameAndDescriptor",     
              ]
"""

# ToDo: implement this (or leave a safe "ignore" mechanism for it)
# class JavaClassInfo(LanguageDependentContainer):
#     ...
""" _fileds = ["id",
               "signature",
               "javaParentInterfaces",
               "classByteCode",
              ]
"""




""" Class description goes here. """

"""The basic dataClay server features for Python Execution Environments.

This module provides the Execution Environment for Python. A basic dataClay
infrastructure is required, mainly:

  - Logic Module
  - Storage Locations
  - [optional] More Execution Environments

Note that this server must be aware of the "local" Storage Location and the
central Logic Module node.
"""
from concurrent import futures
import grpc
import logging
import os
import socket
import sys
import time
import signal
import threading 
import traceback

from dataclay.commonruntime.Runtime import clean_runtime
from dataclay.commonruntime.Settings import settings
from dataclay.communication.grpc.clients.ExecutionEnvGrpcClient import EEClient
from dataclay.communication.grpc.clients.LogicModuleGrpcClient import LMClient
from dataclay.communication.grpc.messages.common.common_messages_pb2 import LANG_PYTHON
from dataclay.communication.grpc.server.ExecutionEnvironmentService import DataServiceEE
from dataclay.util.classloaders import ClassLoader  # Import after DataClayRuntime to avoid circular imports
from dataclay.util.config.CfgExecEnv import set_defaults
from dataclay.executionenv.ExecutionEnvironment import ExecutionEnvironment
from dataclay.commonruntime.Initializer import logger
from dataclay.util import Configuration

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2015 Barcelona Supercomputing Center (BSC-CNS)'

SERVER_TIME_CHECK_SECONDS = 1


class ExecutionEnvironmentSrv(object):
    
    def __init__(self):
        self.execution_environment = None

    def reset_caches(self):
        logger.info("Received SIGHUP --proceeding to reset caches")
        ClassLoader.cached_metaclass_info.clear()
        ClassLoader.cached_metaclasses.clear()
    
    def persist_and_exit(self):
        logger.info("Performing exit hook --persisting files")
        
        self.execution_environment.prepareThread()
        self.execution_environment.get_runtime().stop_gc()
        logger.info("Flushing all objects to disk")
        self.execution_environment.get_runtime().flush_all()
        logger.info("Stopping runtime")
        self.execution_environment.store_ee_info()
        from dataclay.api import finish
        finish()
        clean_runtime()

    def preface_autoregister(self):
        """Perform a pre-initialization of stuff (prior to the autoregister call)."""
        self.execution_environment.prepareThread()
        # logger.info("Preface Autoregister")
    
        # Check if there is an explicit IP for autoregistering
        local_ip = os.getenv("DATASERVICE_HOST", "")
        if not local_ip:
            local_ip = socket.gethostbyname(socket.gethostname())
    
        logger.info("Starting client to LogicModule at %s:%d",
                     settings.logicmodule_host, settings.logicmodule_port)
    
        lm_client = LMClient(settings.logicmodule_host, settings.logicmodule_port)
    
        # Leave the ready client to the LogicModule globally available
        self.execution_environment.get_runtime().ready_clients["@LM"] = lm_client
    
        # logger.info("local_ip %s returned", local_ip)
        return local_ip
    
    def start_autoregister(self, local_ip):
        """Start the autoregister procedure to introduce ourselves to the LogicModule."""
        self.execution_environment.prepareThread()

        logger.info("Start Autoregister with %s local_ip", local_ip)
        lm_client = self.execution_environment.get_runtime().ready_clients["@LM"]
    
        success = False
        retries = 0
        max_retries = Configuration.MAX_RETRY_AUTOREGISTER
        sleep_time = Configuration.RETRY_AUTOREGISTER_TIME / 1000
        execution_environment_id = self.execution_environment.get_execution_environment_id()
        while not success:
            try:
                storage_location_id = lm_client.autoregister_ee(execution_environment_id,
                    settings.dataservice_name,
                    local_ip,
                    settings.dataservice_port,
                    LANG_PYTHON)
            except Exception as e:
                logger.debug("Catched exception of type %s. Message:\n%s", type(e), e)
                if retries > max_retries:
                    logger.warn("Could not create channel, aborting (reraising exception)")
                    raise
                else:
                    logger.info("Could not create channel, retry #%d of %i in %i seconds", retries, max_retries, sleep_time)
                    # TODO: Not Very performing, find a better way
                    time.sleep(sleep_time)
                    retries += 1
            else:
                success = True
    
        logger.info("Current DataService autoregistered. Associated StorageLocationID: %s",
                    storage_location_id)
        settings.storage_id = storage_location_id
        settings.environment_id = execution_environment_id
    
        # Retrieve the storage_location connection data
        storage_location = lm_client.get_storage_location_for_ds(storage_location_id)
    
        logger.debug("StorageLocation data: {name: '%s', hostname: '%s', port: %d}",
                     storage_location.name,
                     storage_location.hostname,
                     storage_location.storageTCPPort)
    
        logger.info("Starting client to StorageLocation {%s} at %s:%d",
                    storage_location_id, storage_location.hostname, storage_location.storageTCPPort)
    
        storage_client = EEClient(storage_location.hostname, storage_location.storageTCPPort)
    
        # Leave the ready client to the Storage Location globally available
        self.execution_environment.get_runtime().ready_clients["@STORAGE"] = storage_client
        storage_client.associate_execution_environment(execution_environment_id)
        
        settings.logicmodule_dc_instance_id = lm_client.get_dataclay_id()
        logger.verbose("DataclayInstanceID is %s, storing client in cache", settings.logicmodule_dc_instance_id)

        self.execution_environment.get_runtime().ready_clients[settings.logicmodule_dc_instance_id] = self.execution_environment.get_runtime().ready_clients["@LM"]

    def start(self):
        """Start the dataClay server (Execution Environment).
    
        Keep in mind that the configuration in both dataClay's global ConfigOptions
        and the server-specific one called ServerConfigOptions should be accurate.
        Furthermore, this function expects that the caller will take care of the
        dataClay library initialization.
    
        This function does not return (by itself), so feel free to spawn it inside
        a greenlet or a subprocess (typical in testing)
        """
        
        set_defaults()

        # Create the deployment folder and add it to the path
        try:
            os.makedirs(settings.deploy_path_source)
        except OSError as e:
            if e.errno != 17:
                # Not the "File exists" expected error, reraise it
                raise
        sys.path.insert(1, settings.deploy_path_source)
        
        self.execution_environment = ExecutionEnvironment(settings.dataservice_name)

        # 0 or undefined should become None, which becomes a default of 5x number of cores (see futures' docs)
        max_workers = Configuration.THREAD_POOL_WORKERS or None

        self.server = grpc.server(futures.ThreadPoolExecutor(max_workers=max_workers),
                                  options=(('grpc.max_send_message_length', -1,),
                                           ('grpc.max_receive_message_length', -1,),
                                           ))
        ee = DataServiceEE(self.execution_environment)
    
        from dataclay.communication.grpc.generated.dataservice import dataservice_pb2_grpc as ds

        ds.add_DataServiceServicer_to_server(ee, self.server)
    
        address = str(settings.server_listen_addr) + ":" + str(settings.server_listen_port)
    
        logger.info("Starting DataServiceEE on %s", address)
        try:
            # ToDo: Better way for start server?
            self.server.add_insecure_port(address)
            self.server.start()
            # ToDo: Check that the server is correctly started
            # ToDo:   -> aka "if the port was in use, fail tremendously and loudly
    
            self.local_ip = self.preface_autoregister()
            self.start_autoregister(self.local_ip)
            ee.ass_client()
            self.running = True 
            signal.signal(signal.SIGINT, self.exit_gracefully_signal)
            signal.signal(signal.SIGTERM, self.exit_gracefully_signal)
            logger.info("Started Python Execution environment on %s", address)
            
            # write state file 
            try:
                f = open("state.txt", "w")
                f.write("READY")
                f.close()
            except:
                logger.info("State file not writable. Skipping file creation.")
                
            try:
                while self.running:
                    time.sleep(SERVER_TIME_CHECK_SECONDS)
            except RuntimeError:
                logger.info("Runtime Error")
        except:
            traceback.print_exc()
        logger.info("** Finished Python Execution Environment on %s", address)

    def exit_gracefully_signal(self, signum, frame):
        self.exit_gracefully()

    def exit_gracefully(self):
        sys.stderr.write("** Exiting gracefully **\n")
        self.persist_and_exit()
        self.server.stop(0)
        self.running = False 
        sys.stderr.write("EXECUTION ENVIRONMENT GRACEFULLY STOPPED\n") 
        
    def get_name(self):
        return settings.dataservice_name




""" Class description goes here. """

""" This is a pool of lockers to manage concurrent actions in shared objects in dataClay """

# Created on Jan 23, 2018

import logging
import threading

__author__ = "dgasull"


class LockerPool(object):
    logger = logging.getLogger('dataclay.LockerPool')

    def __init__(self):
        # Lockers
        self.lockers = dict()

        # Lock for the pool
        self.locker = threading.Lock()
    
    def lock(self, object_id):
        """ 
        Locking design.
        ==============
        If we are using CPython as interpreter, GIL can 'stop' current thread in the middle of a serialization and 
        start sleeping thread in the middle of other action in same object and we can have race conditions or many other
        problems. Therefore we need to lock the objects. 
        
        CPython guarantees that get/put in maps are atomic but not 'get and create if needed'. In our case,
        imagine the code below and two threads located in the lines specified as comments:
        
        if not object_id in self.lockers: #ObjectID is the same for both threads 
            lock = threading.RLock() #Thread 2 is here
            self.lockers.put(object_id, lock) #Thread 1 is here 

        Thread 1 was the first to arrive and find out that there is no locker for the object "he" wants to lock. Therefore,
        Thread 1 creates a new locker. During creation of the Locker, Thread 2 arrives and find no locker.
        Thread 1 put the locker in the ConcurrentHashMap. Thread 2 also creates a locker and overrides Thread 1's locker.
        This is a problem we should solve since both threads are using differents lockers for the same object, loosing the
        concept of lockers.
        
        
        In order to solve that we need to create an atomic set of operations, from now on called "synchronized block of code". 
        We could create a synchronized block based in ObjectID since Threads accessing/creating lockers 
        with same ObjectIDs can have a problem. However, our current implementation does NOT
        guarantee that two ObjectIDs representing same ID (ie with same UUID) are the same instances in memory.
        Therefore we cannot create a synchronized block using ObjectID. We could think about forcing all ObjectIDs to be
        unique ("same-instances") in memory (only one instance of ObjectID representing an object) but it ends in having same 
        synchronization but during the creation of the ObjectID (it can be in any request). Now, adding the synchronized block here is better
        in order to only lock 'if-needed' instead of doing it at creation time.
        
        Then, what should we synchronize? Anything that is 'common' for all threads. In our case, the Locker Pool is unique
        per runtime (server or client). For example:
        
        
        self.lock.acquire() #Lock is created during construction of the LockerPool, "lock = threading.RLock()" It's a reentrant lock!
        if not object_id in self.lockers: 
            lock = threading.RLock() 
            self.lockers.put(object_id, lock)        
        self.lock.release()
    
        
        This would solve our problem. But it's a bottleneck! All threads that wants to lock anything, will end up locking
        the pool. Since our problem is just at 'creation' of lockers we move the synchronized block inside the if and we
        apply the design of 'double-check' of lockers:
        
        if not object_id in self.lockers: 
            self.lock.acquire() # ensure only one locker is created per object
            if not object_id in self.lockers: 
                lock = threading.RLock() 
                self.lockers.put(object_id, lock)  
            self.lock.release()
        else:
            lock = self.lockers.get(object_id)

        
        All threads that try to 'create' lockers are going to have the overhead of synchronizing the locker pool. However,
        it is not our critical path, since in our critical path everything is in cache and lockers exist.
        
        """
        try:
            obj_lock = self.lockers[object_id]
        except KeyError:
            with self.locker:  # ensure only one locker is created per object
                if object_id not in self.lockers:
                    obj_lock = threading.RLock()
                    self.lockers[object_id] = obj_lock
                else:
                    # race condition averted!
                    obj_lock = self.lockers[object_id]

        # Locks the object     
        obj_lock.acquire()
        
    def unlock(self, object_id):
        """
        Unlock object with object ID provided. 
        If the locker does not exist, it means it was already released and cleaned from memory. Since our lockers are 
        Reentrant, there can be many 'unlock' calls unlocking the same locker. If our Memory manager wants to clean a locker,
        it checks if it is unlocked and if so, it removes it. Therefore, just unlock if exists. 
        
        """         
        try:
            self.lockers[object_id].release()
        except KeyError:
            pass
            
    def cleanLockers(self):
        """
        Clean lockers. This function should be called only from Memory manager in order to clean the lockers alive in memory and
        avoid having thousands of 'dead' lockers occupying memory. Since it is only called from ONE thread, it is only possible
        that some thread wants to create a Locker (see lock function) at same time we want to remove it. Therefore, we create
        a synchronized block for that. 
        """
        cleaned_lockers = 0
        
        # We create a copy of the locker_ids, (the key-set) and we remove them if needed.
        locker_ids = list(self.lockers.keys())
        for object_id in locker_ids:
            with self.locker:  # ensure we are not cleaning lockers during lockers creation.
                pass
                # locker = self.lockers.get(object_id)
                # if not locker._
                #    self.lockers.pop(object_id)
                #    cleaned_lockers = cleaned_lockers + 1



"""Basic Synchronization mechanisms."""

from dataclay import dclayMethod


class SequentialConsistencyMixin(object):
    """Simple sequential consistency synchronization mechanism.
    
    This trivial sequential consistency consists on a immediate replication to 
    all slaves. In order to achieve that, all the locations of slave locations
    are iterated and the set operation is performed in them.
    """
    @dclayMethod(attribute="str", value="anything")
    def synchronize(self, attribute, value):
        # This import should be there because this method will be in the stub
        for exeenv_id in self.get_all_locations().keys():
        #if not exeenv_id is master_location:
            self.set_in_backend(exeenv_id, attribute, value)
            
    @dclayMethod(attribute="str", value="anything")
    def synchronize_federated(self, attribute, value):
        for dataclay_id in self.get_federation_targets():
            self.set_in_dataclay_instance(dataclay_id, attribute, [value])
        dataclay_id = self.get_federation_source()
        if dataclay_id is not None:
            self.set_in_dataclay_instance(dataclay_id, attribute, [value])




""" Class description goes here. """

from dataclay.util.MgrObject import ManagementObject


class AccessedImplementation(ManagementObject):
    _fields = ["id",  # sql id, internals
               "namespace",
               "className",
               "opSignature",
               "implPosition",
               ]

    _internal_fields = ["implementationID",
                        ]




""" Class description goes here. """

import logging
import re
import six
import traceback
if six.PY2:
    import cPickle as pickle
elif six.PY3:
    import _pickle as pickle
    
from dataclay.commonruntime.Initializer import size_tracking
from dataclay.serialization.python.DataClayPythonWrapper import DataClayPythonWrapper
from dataclay.serialization.python.lang.BooleanWrapper import BooleanWrapper
from dataclay.serialization.python.lang.IntegerWrapper import IntegerWrapper
from dataclay.serialization.python.lang.StringWrapper import StringWrapper

import six

logger = logging.getLogger(__name__)

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2015 Barcelona Supercomputing Center (BSC-CNS)'


def safe_wait_if_compss_future(potential_future):
    """Safe approach to COMPSs Futures: compss_wait_on if that is a Future.

    COMPSs may use Future objects, and this function returns the real object
    (after a call to compss_wait_on to the COMPSs API). If the objects was not
    a Future, return the object itself.

    :param potential_future: May be a PyCOMPSs Future instance
    :return: NOT a Future instance. Redirect potential_future
    """
    real_type = type(potential_future)

    if real_type.__name__ == "Future" and \
            real_type.__module__ == "pycompss.runtime.binding":
        from pycompss.api.api import compss_wait_on

        logger.info("Received a `Future` PyCOMPSs object, waiting for the real object...")
        param = compss_wait_on(potential_future)
        real_type = type(param)
        logger.info("Using the parameter: %r (type: %s)",
                    param, real_type)
    else:
        param = potential_future

    return param


class PyTypeWildcardWrapper(DataClayPythonWrapper):
    """Generic catch-all for Python types (including custom-signature binary types)."""
    __slots__ = ("_signature", "_pickle_fallback")

    PYTHON_PREFIX = 'python.'

    # Note that this regex does not have guarantees on matching <> or [] (so <] will be a valid group,
    # as so will be its [> counterpart). But... that seems a user problem.
    # Also: deep and esoteric nesting is not supported (regex should be thrown away and a sane
    # markup reader used instead).
    SEQUENCE_REGEX = re.compile(r'(?P<base_type>(list)|(tuple)|(set))\s*(?:[<\[]\s*(?P<subtype>.*?)\s*[>\]])?\s*$')
    MAPPING_REGEX = re.compile(r'(?P<base_type>dict)\s*(?:[<\[]\s*(?P<keytype>.*?)\s*,\s*(?P<valuetype>.*?)\s*[>\]])?\s*$')
    STR_SIGNATURE = 'str'
    UNICODE_SIGNATURE = 'unicode'
    STORAGEOBJECT_SIGNATURE = 'storageobject'
    ANYTHING_SIGNATURE = 'anything'
    NUMPY_SIGNATURE = 'numpy'

    def __init__(self, signature, pickle_fallback=False):
        # TODO make some checks, and raise InvalidPythonSignature otherwise
        self._signature = signature
        self._pickle_fallback = pickle_fallback

    def read(self, io_file):
        from dataclay.util.management.classmgr.Utils import serialization_types
        try:
            return serialization_types[self._signature].read(io_file)
        except KeyError:
            pass

        # numpy have their own special ultra-fast serialization
        if self._signature.startswith(self.NUMPY_SIGNATURE):
            import numpy as np
            # Ignoring field size, as numpy is selfcontained in that matter
            _ = IntegerWrapper(32).read(io_file)
            return np.load(io_file, allow_pickle=False)

        # anything is also a special case, also all its alias
        if self._signature == self.ANYTHING_SIGNATURE or \
                self._signature == self.STORAGEOBJECT_SIGNATURE:
            field_size = IntegerWrapper(32).read(io_file)
            logger.debug("Deserializing DataClayObject from pickle")

            return pickle.loads(io_file.read(field_size))

        # Everything shoulda be a python type...
        if not self._signature.startswith(self.PYTHON_PREFIX):
            # ... except the fallbacks (mostly for subtypes like lists of persistent objects)
            # TODO: Check pickle fallback or ignore it completely
            field_size = IntegerWrapper(32).read(io_file)
            return pickle.loads(io_file.read(field_size))

        subtype = self._signature[len(self.PYTHON_PREFIX):]

        sequence_match = self.SEQUENCE_REGEX.match(subtype)
        mapping_match = self.MAPPING_REGEX.match(subtype)

        if sequence_match:
            gd = sequence_match.groupdict()
            logger.debug("Deserializing a Python Sequence with the following match: %s", gd)

            if gd["subtype"]:
                instances_type = PyTypeWildcardWrapper(gd["subtype"], pickle_fallback=True)
            else:  # list without subtypes information
                instances_type = PyTypeWildcardWrapper(self.ANYTHING_SIGNATURE)

            ret = list()
            size = IntegerWrapper(32).read(io_file)
            logger.debug("### READ SIZE OF SEQUENCE MATCH: %i", size)

            for i in range(size):
                if BooleanWrapper().read(io_file):
                    ret.append(instances_type.read(io_file))
                else:
                    ret.append(None)
            
            if gd["base_type"] == "tuple":
                logger.debug("Returning deserialized Python tuple")
                return tuple(ret)
            else:
                logger.debug("Returning deserialized Python list")
                return ret

        elif mapping_match:
            gd = mapping_match.groupdict()
            logger.debug("Deserializing a Python mapping with the following match: %s", gd)

            if gd["keytype"] and gd["valuetype"]:
                key_type = PyTypeWildcardWrapper(gd["keytype"], pickle_fallback=True)
                value_type = PyTypeWildcardWrapper(gd["valuetype"], pickle_fallback=True)
            else:
                # dict without subtypes information
                key_type = PyTypeWildcardWrapper(self.ANYTHING_SIGNATURE)
                value_type = PyTypeWildcardWrapper(self.ANYTHING_SIGNATURE)

            ret = dict()
            size = IntegerWrapper(32).read(io_file)
            for i in range(size):
                if BooleanWrapper().read(io_file):
                    key = key_type.read(io_file)
                else:
                    key = None

                if BooleanWrapper().read(io_file):
                    ret[key] = value_type.read(io_file)
                else:
                    ret[key] = None
            logger.debug("Returning deserialized Python map")
            return ret

        elif subtype == self.STR_SIGNATURE:
            if six.PY2:
                return StringWrapper('binary').read(io_file)
            elif six.PY3:
                return StringWrapper('utf-8').read(io_file)
        elif subtype == self.UNICODE_SIGNATURE:
            return StringWrapper('utf-16').read(io_file)
        else:
            raise NotImplementedError("Python types supported at the moment: "
                                      "list and mappings (but not `%s`), sorry" % subtype)

    def write(self, io_file, value):
        value = safe_wait_if_compss_future(value)

        from dataclay.util.management.classmgr.Utils import serialization_types
        try:
            serialization_types[self._signature].write(io_file, value)
            return
        except KeyError:
            pass

        # numpy have their own special ultra-fast serialization
        if self._signature.startswith(self.NUMPY_SIGNATURE):
            import numpy as np
            with size_tracking(io_file):
                np.save(io_file, value)
            return

        # anything is also a special case, also all its alias
        if self._signature == self.ANYTHING_SIGNATURE or \
                self._signature == self.STORAGEOBJECT_SIGNATURE:
            s = pickle.dumps(value, protocol=-1)
            IntegerWrapper(32).write(io_file, len(s))
            io_file.write(s)
            return

        # Everything shoulda be a python type...
        if not self._signature.startswith(self.PYTHON_PREFIX):
            # ... except the fallbacks (mostly for subtypes like lists of persistent objects)
            # TODO: Check pickle fallback or ignore it completely

            s = pickle.dumps(value, protocol=-1)
            IntegerWrapper(32).write(io_file, len(s))
            io_file.write(s)
            return

        # Now everything must be a python type
        assert self._signature.startswith(self.PYTHON_PREFIX), \
            "Signature for Python types is expected to start with " \
            "'python'. Found signature: %s" % self._signature

        subtype = self._signature[len(self.PYTHON_PREFIX):]

        sequence_match = self.SEQUENCE_REGEX.match(subtype)
        mapping_match = self.MAPPING_REGEX.match(subtype)

        if sequence_match:
            gd = sequence_match.groupdict()
            logger.debug("Serializing a Python Sequence with the following match: %s", gd)

            if gd["subtype"]:
                instances_type = PyTypeWildcardWrapper(gd["subtype"], pickle_fallback=True)
            else:  # list without subtypes information
                instances_type = PyTypeWildcardWrapper(self.ANYTHING_SIGNATURE)

            IntegerWrapper(32).write(io_file, len(value))
            for elem in value:
                if elem is None:
                    BooleanWrapper().write(io_file, False)
                else:
                    BooleanWrapper().write(io_file, True)
                    instances_type.write(io_file, elem)

        elif mapping_match:
            gd = mapping_match.groupdict()
            logger.debug("Serializing a Python Mapping with the following match: %s", gd)

            if gd["keytype"] and gd["valuetype"]:
                key_type = PyTypeWildcardWrapper(gd["keytype"], pickle_fallback=True)
                value_type = PyTypeWildcardWrapper(gd["valuetype"], pickle_fallback=True)
            else:  # dict without subtypes information
                key_type = PyTypeWildcardWrapper(self.ANYTHING_SIGNATURE)
                value_type = PyTypeWildcardWrapper(self.ANYTHING_SIGNATURE)

            IntegerWrapper(32).write(io_file, len(value))

            for k, v in value.items():
                if k is None:
                    BooleanWrapper().write(io_file, False)
                else:
                    BooleanWrapper().write(io_file, True)
                    key_type.write(io_file, k)

                if v is None:
                    BooleanWrapper().write(io_file, False)
                else:
                    # ToDo remove this when COMPSs behaves correctly with compss_wait_on(dict_instance)
                    v = safe_wait_if_compss_future(v)

                    BooleanWrapper().write(io_file, True)
                    value_type.write(io_file, v)

        elif subtype == self.STR_SIGNATURE:
            if six.PY2:
                StringWrapper('utf-8').write(io_file, value)
            elif six.PY3:
                StringWrapper('binary').write(io_file, value)
        elif subtype == self.UNICODE_SIGNATURE:
            StringWrapper('utf-16').write(io_file, value)
        else:
            raise NotImplementedError("Python types supported at the moment: "
                                      "list and mappings (but not `%s`), sorry" % subtype)




from __future__ import print_function
""" Class description goes here. """

"""Entrypoint for tool execution.

This module can be executed with `python -m dataclay.tool <parameters>`.
"""
import sys
from . import functions
import six

if len(sys.argv) < 2:
    # We need at least a parameter
    print("dataclay.tool requires at least the function parameter", file=sys.stderr)
    exit(1)

if six.PY2:
    func_name = sys.argv[1]
    func = getattr(functions, func_name)

elif six.PY3:
    __name__ = sys.argv[1]
    func = getattr(functions, __name__)

if not func:
    print("Unknown dataclay.tool function '%s'" % func_name, file=sys.stderr)
    exit(2)

func(*sys.argv[2:])




""" Class description goes here. """

from dataclay.serialization.python.DataClayPythonWrapper import DataClayPythonWrapper
from six import int2byte

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2015 Barcelona Supercomputing Center (BSC-CNS)'

class VLQIntegerWrapper(DataClayPythonWrapper):
    """Variable Length Quantity."""
    __slots__ = ()

    def __init__(self):
        pass

    def read(self, io_file):
        value = 0
        while True:
            # Read one byte and build
            b = ord(io_file.read(1))
            value = (value << 7) + (b & 0x7F)

            # If the continuation bit (MSB) is zero, we are finished
            if (b & 0x80) == 0:
                return value

    # FIXME: Check with Python 2.7
    def write(self, io_file, value):
        if value == 0:
            io_file.write(b'\x00')
            return
        values = []
        while value > 0:
            # Put the 7-bit values into the list
            values.append(value & 0x7F)
            value >>= 7

        # Values with continuation bit activated
        for b in reversed(values[1:]):
            io_file.write(int2byte(0x80 | b))

        # Last value, no continuation bit
        io_file.write(int2byte(values[0]))




""" Class description goes here. """

import os
import logging
from dataclay.commonruntime.Settings import settings

__author__ = 'Alex Barcelo <alex.barcelo@bsc.es>'
__copyright__ = '2015 Barcelona Supercomputing Center (BSC-CNS)'
logger = logging.getLogger(__name__)


def set_defaults():
    # extrae
    settings.tracing_enabled = False
    settings.extrae_starting_task_id = 0

    # Firewall it if this listen address is not advisable
    settings.server_listen_addr = "0.0.0.0"

    port_str = os.getenv("DATASERVICE_PYTHON_PORT_TCP", "6867")
    settings.server_listen_port = int(port_str)

    settings.pool_size = 4
    settings.interpreter_pool = None

    settings.cached_objects_size = 100
    settings.cached_classes_size = 1000
    settings.cached_metaclasses_size = 1000
    settings.cached_metaclass_info_size = 1000

    settings.cache_on_deploy = True

    # ToDo: This is a dangerous default, used in Docker but... useless everywhere else.
        
    deploy_path = os.getenv("DEPLOY_PATH", "/dataclay/deploy")
    settings.deploy_path = deploy_path
    settings.deploy_path_source = os.getenv("DEPLOY_PATH_SRC",
                                            os.path.join(deploy_path, "source"))
    # Retain the IDs associated to this server
    settings.storage_id = None  # The Java, which is both StorageLocation and Java ExecutionEnvironment
    settings.environment_id = None  # This is (should be) ourselves

    ###########################################################################
    # The following is the "published" information for the autoregister RPC
    settings.dataservice_name = os.getenv("DATASERVICE_NAME")
    # If there are no firewalls or strange stuff with ports,
    # the following is expected to match DATASERVICE_PYTHON_PORT_TCP
    # (the port the server listens to)
    settings.dataservice_port = int(os.getenv("DATASERVICE_PYTHON_PORT_TCP", port_str))
    #                                                                         #
    ###########################################################################

    settings.logicmodule_host = os.getenv("LOGICMODULE_HOST", "127.0.0.1")
    settings.logicmodule_rmiport = int(os.getenv("LOGICMODULE_PORT_RMI", "1024"))
    settings.logicmodule_port = int(os.getenv("LOGICMODULE_PORT_TCP", "1034"))

    # Setting defaults is considered loading settings
    settings.loaded = True




""" Class description goes here. """

from .Type import Type


# Deleted name from fields
class UserType(Type):
    _fields = [
               "namespace",
               ]

    _internal_fields = ["classID",
                        ]

